
///////////////////////////////////////////////////
//                                               //
//   AcedMemory 1.09                             //
//                                               //
//     /    //
//                                               //
//   mailto: acedutils@yandex.ru                 //
//                                               //
///////////////////////////////////////////////////

unit AcedMemory;

(*
  AcedMemory     , 
     (RTL)    GetMem, FreeMem
   ReallocMem.        uses 
  . :

  program Project1;

  uses
    AcedMemory in 'AcedMemory.pas',
    Forms,
    ...

  {$R *.RES}

  begin
    Application.Initialize;
    Application.Title := 'Project1';
    Application.CreateForm(TMainForm, MainForm);
    ...
    Application.Run;
  end.
*)

interface

{       . 
     : GetMem/FreeMem. }

function G_GetMem(Size: Integer): Pointer;
function G_FreeMem(P: Pointer): Integer;

{ G_GetMemSize    ,   
   P. }

function G_GetMemSize(P: Pointer): Integer;

{ G_GetTotalMemAllocated    , 
  . }

function G_GetTotalMemAllocated: LongWord;

{ G_GetTotalMemCommitted     ,
          . }

function G_GetTotalMemCommitted: LongWord;

implementation

uses Windows, SysUtils, AcedBinary, AcedConsts;

const
  mbfManagerInstalled     = $0001;
  mbfFreeRegionExists     = $0002;
  mbfFreeBlock8Exists     = $0004;
  mbfFreeBlock16Exists    = $0008;
  mbfFreeBlock32Exists    = $0010;
  mbfFreeBlock64Exists    = $0020;
  mbfFreeBlock128Exists   = $0040;
  mbfFreeBlock256Exists   = $0080;
  mbfFreeBlock512Exists   = $0100;
  mbfFreeBlock1024Exists  = $0200;
  mbfFreeBlock2kExists    = $0400;
  mbfFreeBlock4kExists    = $0800;
  mbfFreeBlock8kExists    = $1000;
  mbfFreeBlock16kExists   = $2000;
  mbfFreeBlock32kExists   = $4000;
  mbfFreeBlock64kExists   = $8000;

  winHEAP_NO_SERIALIZE = 1;
  winHEAP_GENERATE_EXCEPTIONS = 4;
  winHEAP_ZERO_MEMORY = 8;

  CountDelta = 8;

type
  PRegion = ^TRegion;
  TRegion = record
    BaseAddr: LongWord;
    BlockMask: LongWord;
    ItemSize0: LongWord;
    ItemSize1: LongWord;
    ItemSize2: LongWord;
    ItemSize3: LongWord;
  end;

  PBlock8 = ^TBlock8;
  TBlock8 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..255] of LongWord;
    Region: PRegion;
  end;

  PBlock16 = ^TBlock16;
  TBlock16 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..127] of LongWord;
    Region: PRegion;
  end;

  PBlock32 = ^TBlock32;
  TBlock32 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..63] of LongWord;
    Region: PRegion;
  end;

  PBlock64 = ^TBlock64;
  TBlock64 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..31] of LongWord;
    Region: PRegion;
  end;

  PBlock128 = ^TBlock128;
  TBlock128 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..15] of LongWord;
    Region: PRegion;
  end;

  PBlock256 = ^TBlock256;
  TBlock256 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..7] of LongWord;
    Region: PRegion;
  end;

  PBlock512 = ^TBlock512;
  TBlock512 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..3] of LongWord;
    Region: PRegion;
  end;

  PBlock1024 = ^TBlock1024;
  TBlock1024 = record
    FirstFree: Integer;
    FreeCount: Integer;
    BaseAddr: LongWord;
    BitArray: array[0..1] of LongWord;
    Region: PRegion;
  end;

  PBlock2k = ^TBlock2k;
  TBlock2k = record
    BaseAddr: LongWord;
    BitArray: LongWord;
    Region: PRegion;
  end;

  PBlock4k = ^TBlock4k;
  TBlock4k = record
    BaseAddr: LongWord;
    BitArray: LongWord;
    Region: PRegion;
  end;

  PBlock8k = ^TBlock8k;
  TBlock8k = record
    BaseAddr: LongWord;
    BitArray: LongWord;
    Region: PRegion;
  end;

  PBlock16k = ^TBlock16k;
  TBlock16k = record
    BaseAddr: LongWord;
    BitArray: LongWord;
    Region: PRegion;
  end;

  PBlock32k = ^TBlock32k;
  TBlock32k = record
    BaseAddr: LongWord;
    BitArray: LongWord;
    Region: PRegion;
  end;

  PBlock64k = ^TBlock64k;
  TBlock64k = record
    BaseAddr: LongWord;
    Region: PRegion;
  end;

  PRegionList = ^TRegionList;
  TRegionList = array[0..DWordListUpperLimit] of PRegion;

  PBlock8List = ^TBlock8List;
  TBlock8List = array[0..DWordListUpperLimit] of PBlock8;

  PBlock16List = ^TBlock16List;
  TBlock16List = array[0..DWordListUpperLimit] of PBlock16;

  PBlock32List = ^TBlock32List;
  TBlock32List = array[0..DWordListUpperLimit] of PBlock32;

  PBlock64List = ^TBlock64List;
  TBlock64List = array[0..DWordListUpperLimit] of PBlock64;

  PBlock128List = ^TBlock128List;
  TBlock128List = array[0..DWordListUpperLimit] of PBlock128;

  PBlock256List = ^TBlock256List;
  TBlock256List = array[0..DWordListUpperLimit] of PBlock256;

  PBlock512List = ^TBlock512List;
  TBlock512List = array[0..DWordListUpperLimit] of PBlock512;

  PBlock1024List = ^TBlock1024List;
  TBlock1024List = array[0..DWordListUpperLimit] of PBlock1024;

  PBlock2kList = ^TBlock2kList;
  TBlock2kList = array[0..DWordListUpperLimit] of PBlock2k;

  PBlock4kList = ^TBlock4kList;
  TBlock4kList = array[0..DWordListUpperLimit] of PBlock4k;

  PBlock8kList = ^TBlock8kList;
  TBlock8kList = array[0..DWordListUpperLimit] of PBlock8k;

  PBlock16kList = ^TBlock16kList;
  TBlock16kList = array[0..DWordListUpperLimit] of PBlock16k;

  PBlock32kList = ^TBlock32kList;
  TBlock32kList = array[0..DWordListUpperLimit] of PBlock32k;

  PBlock64kList = ^TBlock64kList;
  TBlock64kList = array[0..DWordListUpperLimit] of PBlock64k;

  PBigBlockList = ^TBigBlockList;
  TBigBlockList = array[0..DWordListUpperLimit] of Integer;

var
  MemFlags: LongWord = 0;
  TotalMemAllocated: LongWord;
  BlockIndex: Integer;
  CriticalSection: _RTL_CRITICAL_SECTION;
  OldMemoryManager: TMemoryManager;
  Heap: THandle;

  RegionCount: Integer = 0;
  Regions: PRegionList;
  RegionCapacity: Integer;

  Block8Count: Integer = 0;
  Blocks8: PBlock8List;
  Block8Capacity: Integer;

  Block16Count: Integer = 0;
  Blocks16: PBlock16List;
  Block16Capacity: Integer;

  Block32Count: Integer = 0;
  Blocks32: PBlock32List;
  Block32Capacity: Integer;

  Block64Count: Integer = 0;
  Blocks64: PBlock64List;
  Block64Capacity: Integer;

  Block128Count: Integer = 0;
  Blocks128: PBlock128List;
  Block128Capacity: Integer;

  Block256Count: Integer = 0;
  Blocks256: PBlock256List;
  Block256Capacity: Integer;

  Block512Count: Integer = 0;
  Blocks512: PBlock512List;
  Block512Capacity: Integer;

  Block1024Count: Integer = 0;
  Blocks1024: PBlock1024List;
  Block1024Capacity: Integer;

  Block2kCount: Integer = 0;
  Blocks2k: PBlock2kList;
  Block2kCapacity: Integer;

  Block4kCount: Integer = 0;
  Blocks4k: PBlock4kList;
  Block4kCapacity: Integer;

  Block8kCount: Integer = 0;
  Blocks8k: PBlock8kList;
  Block8kCapacity: Integer;

  Block16kCount: Integer = 0;
  Blocks16k: PBlock16kList;
  Block16kCapacity: Integer;

  Block32kCount: Integer = 0;
  Blocks32k: PBlock32kList;
  Block32kCapacity: Integer;

  Block64kCount: Integer = 0;
  Blocks64k: PBlock64kList;
  Block64kCapacity: Integer;

  BigBlockCount: Integer = 0;
  BigBlockAddrs: PBigBlockList;
  BigBlockSizes: PBigBlockList;
  BigBlockCapacity: Integer;

function G_GetRegion: PRegion;
var
  I: Integer;
  M: LongWord;
  NewRegions: PRegionList;
begin
  for I := 0 to RegionCount - 1 do
  begin
    Result := Regions^[I];
    M := Result.BlockMask;
    if M <> $FFFFFFFF then
    begin
      BlockIndex := G_FreeBitScanForward32(M);
      Result.BlockMask := G_BitSet32(M, BlockIndex);
      Exit;
    end;
  end;
  if MemFlags and mbfFreeRegionExists <> 0 then
  begin
    BlockIndex := 0;
    Result := Regions^[RegionCount];
    Result.BlockMask := 1;
    Inc(RegionCount);
    MemFlags := MemFlags and not mbfFreeRegionExists;
    Exit;
  end;
  if RegionCount = RegionCapacity then
  begin
    Inc(RegionCapacity, CountDelta);
    NewRegions := HeapAlloc(Heap, 0, RegionCapacity * SizeOf(Pointer));
    G_CopyLongs(Regions, NewRegions, RegionCapacity - CountDelta);
    HeapFree(Heap, 0, Regions);
    Regions := NewRegions;
    for I := CountDelta downto 1 do
      Regions^[RegionCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TRegion));
  end;
  BlockIndex := 0;
  Result := Regions^[RegionCount];
  with Result^ do
  begin
    BlockMask := 1;
    BaseAddr := LongWord(VirtualAlloc(nil, $200000,
      MEM_RESERVE or MEM_COMMIT or MEM_TOP_DOWN, PAGE_READWRITE));
    if BaseAddr <> 0 then
      Inc(RegionCount)
    else
      Result := nil;
  end;
end;

function G_GetMem8: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks8: PBlock8List;
begin
  Result := nil;
  for I := 0 to Block8Count - 1 do
    with Blocks8^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 3) + BaseAddr);
        Inc(TotalMemAllocated, 8);
        Exit;
      end;
  if MemFlags and mbfFreeBlock8Exists <> 0 then
    with Blocks8^[Block8Count]^ do
    begin
      FreeCount := 8191;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block8Count);
      MemFlags := MemFlags and not mbfFreeBlock8Exists;
      Inc(TotalMemAllocated, 8);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitReset32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitReset32(ItemSize3, BlockIndex);
    end;
    if Block8Count = Block8Capacity then
    begin
      Inc(Block8Capacity, CountDelta);
      NewBlocks8 := HeapAlloc(Heap, 0, Block8Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks8, NewBlocks8, Block8Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks8);
      Blocks8 := NewBlocks8;
      for I := CountDelta downto 1 do
        Blocks8^[Block8Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock8));
    end;
    with Blocks8^[Block8Count]^ do
    begin
      FreeCount := 8191;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 8);
    Inc(Block8Count);
  end;
end;

function G_GetMem16: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks16: PBlock16List;
begin
  Result := nil;
  for I := 0 to Block16Count - 1 do
    with Blocks16^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 4) + BaseAddr);
        Inc(TotalMemAllocated, 16);
        Exit;
      end;
  if MemFlags and mbfFreeBlock16Exists <> 0 then
    with Blocks16^[Block16Count]^ do
    begin
      FreeCount := 4095;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block16Count);
      MemFlags := MemFlags and not mbfFreeBlock16Exists;
      Inc(TotalMemAllocated, 16);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitReset32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitReset32(ItemSize3, BlockIndex);
    end;
    if Block16Count = Block16Capacity then
    begin
      Inc(Block16Capacity, CountDelta);
      NewBlocks16 := HeapAlloc(Heap, 0, Block16Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks16, NewBlocks16, Block16Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks16);
      Blocks16 := NewBlocks16;
      for I := CountDelta downto 1 do
        Blocks16^[Block16Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock16));
    end;
    with Blocks16^[Block16Count]^ do
    begin
      FreeCount := 4095;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 16);
    Inc(Block16Count);
  end;
end;

function G_GetMem32: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks32: PBlock32List;
begin
  Result := nil;
  for I := 0 to Block32Count - 1 do
    with Blocks32^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 5) + BaseAddr);
        Inc(TotalMemAllocated, 32);
        Exit;
      end;
  if MemFlags and mbfFreeBlock32Exists <> 0 then
    with Blocks32^[Block32Count]^ do
    begin
      FreeCount := 2047;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block32Count);
      MemFlags := MemFlags and not mbfFreeBlock32Exists;
      Inc(TotalMemAllocated, 32);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitReset32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitReset32(ItemSize3, BlockIndex);
    end;
    if Block32Count = Block32Capacity then
    begin
      Inc(Block32Capacity, CountDelta);
      NewBlocks32 := HeapAlloc(Heap, 0, Block32Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks32, NewBlocks32, Block32Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks32);
      Blocks32 := NewBlocks32;
      for I := CountDelta downto 1 do
        Blocks32^[Block32Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock32));
    end;
    with Blocks32^[Block32Count]^ do
    begin
      FreeCount := 2047;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 32);
    Inc(Block32Count);
  end;
end;

function G_GetMem64: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks64: PBlock64List;
begin
  Result := nil;
  for I := 0 to Block64Count - 1 do
    with Blocks64^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 6) + BaseAddr);
        Inc(TotalMemAllocated, 64);
        Exit;
      end;
  if MemFlags and mbfFreeBlock64Exists <> 0 then
    with Blocks64^[Block64Count]^ do
    begin
      FreeCount := 1023;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block64Count);
      MemFlags := MemFlags and not mbfFreeBlock64Exists;
      Inc(TotalMemAllocated, 64);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitReset32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitReset32(ItemSize3, BlockIndex);
    end;
    if Block64Count = Block64Capacity then
    begin
      Inc(Block64Capacity, CountDelta);
      NewBlocks64 := HeapAlloc(Heap, 0, Block64Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks64, NewBlocks64, Block64Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks64);
      Blocks64 := NewBlocks64;
      for I := CountDelta downto 1 do
        Blocks64^[Block64Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock64));
    end;
    with Blocks64^[Block64Count]^ do
    begin
      FreeCount := 1023;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 64);
    Inc(Block64Count);
  end;
end;

function G_GetMem128: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks128: PBlock128List;
begin
  Result := nil;
  for I := 0 to Block128Count - 1 do
    with Blocks128^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 7) + BaseAddr);
        Inc(TotalMemAllocated, 128);
        Exit;
      end;
  if MemFlags and mbfFreeBlock128Exists <> 0 then
    with Blocks128^[Block128Count]^ do
    begin
      FreeCount := 511;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block128Count);
      MemFlags := MemFlags and not mbfFreeBlock128Exists;
      Inc(TotalMemAllocated, 128);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitReset32(ItemSize3, BlockIndex);
    end;
    if Block128Count = Block128Capacity then
    begin
      Inc(Block128Capacity, CountDelta);
      NewBlocks128 := HeapAlloc(Heap, 0, Block128Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks128, NewBlocks128, Block128Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks128);
      Blocks128 := NewBlocks128;
      for I := CountDelta downto 1 do
        Blocks128^[Block128Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock128));
    end;
    with Blocks128^[Block128Count]^ do
    begin
      FreeCount := 511;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 128);
    Inc(Block128Count);
  end;
end;

function G_GetMem256: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks256: PBlock256List;
begin
  Result := nil;
  for I := 0 to Block256Count - 1 do
    with Blocks256^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 8) + BaseAddr);
        Inc(TotalMemAllocated, 256);
        Exit;
      end;
  if MemFlags and mbfFreeBlock256Exists <> 0 then
    with Blocks256^[Block256Count]^ do
    begin
      FreeCount := 255;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block256Count);
      MemFlags := MemFlags and not mbfFreeBlock256Exists;
      Inc(TotalMemAllocated, 256);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitReset32(ItemSize3, BlockIndex);
    end;
    if Block256Count = Block256Capacity then
    begin
      Inc(Block256Capacity, CountDelta);
      NewBlocks256 := HeapAlloc(Heap, 0, Block256Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks256, NewBlocks256, Block256Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks256);
      Blocks256 := NewBlocks256;
      for I := CountDelta downto 1 do
        Blocks256^[Block256Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock256));
    end;
    with Blocks256^[Block256Count]^ do
    begin
      FreeCount := 255;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 256);
    Inc(Block256Count);
  end;
end;

function G_GetMem512: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks512: PBlock512List;
begin
  Result := nil;
  for I := 0 to Block512Count - 1 do
    with Blocks512^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 9) + BaseAddr);
        Inc(TotalMemAllocated, 512);
        Exit;
      end;
  if MemFlags and mbfFreeBlock512Exists <> 0 then
    with Blocks512^[Block512Count]^ do
    begin
      FreeCount := 127;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block512Count);
      MemFlags := MemFlags and not mbfFreeBlock512Exists;
      Inc(TotalMemAllocated, 512);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitReset32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitReset32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block512Count = Block512Capacity then
    begin
      Inc(Block512Capacity, CountDelta);
      NewBlocks512 := HeapAlloc(Heap, 0, Block512Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks512, NewBlocks512, Block512Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks512);
      Blocks512 := NewBlocks512;
      for I := CountDelta downto 1 do
        Blocks512^[Block512Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock512));
    end;
    with Blocks512^[Block512Count]^ do
    begin
      FreeCount := 127;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 512);
    Inc(Block512Count);
  end;
end;

function G_GetMem1024: Pointer;
var
  I, J, Index: Integer;
  R: PRegion;
  NewBlocks1024: PBlock1024List;
begin
  Result := nil;
  for I := 0 to Block1024Count - 1 do
    with Blocks1024^[I]^ do
      if FreeCount > 0 then
      begin
        J := FirstFree;
        Index := (J shl 5) + G_FreeBitScanForward32(BitArray[J]);
        Dec(FreeCount);
        G_BitSet(@BitArray, Index);
        if FreeCount > 0 then
        begin
          while BitArray[J] = $FFFFFFFF do
            Inc(J);
          FirstFree := J;
        end else
          FirstFree := -1;
        Result := Pointer(LongWord(Index shl 10) + BaseAddr);
        Inc(TotalMemAllocated, 1024);
        Exit;
      end;
  if MemFlags and mbfFreeBlock1024Exists <> 0 then
    with Blocks1024^[Block1024Count]^ do
    begin
      FreeCount := 63;
      BitArray[0] := 1;
      Result := Pointer(BaseAddr);
      Inc(Block1024Count);
      MemFlags := MemFlags and not mbfFreeBlock1024Exists;
      Inc(TotalMemAllocated, 1024);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitReset32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitReset32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block1024Count = Block1024Capacity then
    begin
      Inc(Block1024Capacity, CountDelta);
      NewBlocks1024 := HeapAlloc(Heap, 0, Block1024Capacity * SizeOf(Pointer));
      G_CopyLongs(Blocks1024, NewBlocks1024, Block1024Capacity - CountDelta);
      HeapFree(Heap, 0, Blocks1024);
      Blocks1024 := NewBlocks1024;
      for I := CountDelta downto 1 do
        Blocks1024^[Block1024Capacity - I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock1024));
    end;
    with Blocks1024^[Block1024Count]^ do
    begin
      FreeCount := 63;
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray[0] := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, 1024);
    Inc(Block1024Count);
  end;
end;

function G_GetMem2k: Pointer;
var
  I, Index: Integer;
  R: PRegion;
  NewBlocks2k: PBlock2kList;
begin
  Result := nil;
  for I := 0 to Block2kCount - 1 do
    with Blocks2k^[I]^ do
      if BitArray <> $FFFFFFFF then
      begin
        Index := G_FreeBitScanForward32(BitArray);
        BitArray := G_BitSet32(BitArray, Index);
        Result := Pointer(LongWord(Index shl 11) + BaseAddr);
        Inc(TotalMemAllocated, $800);
        Exit;
      end;
  if MemFlags and mbfFreeBlock2kExists <> 0 then
    with Blocks2k^[Block2kCount]^ do
    begin
      BitArray := 1;
      Result := Pointer(BaseAddr);
      Inc(Block2kCount);
      MemFlags := MemFlags and not mbfFreeBlock2kExists;
      Inc(TotalMemAllocated, $800);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitReset32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block2kCount = Block2kCapacity then
    begin
      Inc(Block2kCapacity, CountDelta);
      NewBlocks2k := HeapAlloc(Heap, 0, Block2kCapacity * SizeOf(Pointer));
      G_CopyLongs(Blocks2k, NewBlocks2k, Block2kCapacity - CountDelta);
      HeapFree(Heap, 0, Blocks2k);
      Blocks2k := NewBlocks2k;
      for I := CountDelta downto 1 do
        Blocks2k^[Block2kCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TBlock2k));
    end;
    with Blocks2k^[Block2kCount]^ do
    begin
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, $800);
    Inc(Block2kCount);
  end;
end;

function G_GetMem4k: Pointer;
var
  I, Index: Integer;
  R: PRegion;
  NewBlocks4k: PBlock4kList;
begin
  Result := nil;
  for I := 0 to Block4kCount - 1 do
    with Blocks4k^[I]^ do
      if BitArray <> $FFFF then
      begin
        Index := G_FreeBitScanForward32(BitArray);
        BitArray := G_BitSet32(BitArray, Index);
        Result := Pointer(LongWord(Index shl 12) + BaseAddr);
        Inc(TotalMemAllocated, $1000);
        Exit;
      end;
  if MemFlags and mbfFreeBlock4kExists <> 0 then
    with Blocks4k^[Block4kCount]^ do
    begin
      BitArray := 1;
      Result := Pointer(BaseAddr);
      Inc(Block4kCount);
      MemFlags := MemFlags and not mbfFreeBlock4kExists;
      Inc(TotalMemAllocated, $1000);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitReset32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block4kCount = Block4kCapacity then
    begin
      Inc(Block4kCapacity, CountDelta);
      NewBlocks4k := HeapAlloc(Heap, 0, Block4kCapacity * SizeOf(Pointer));
      G_CopyLongs(Blocks4k, NewBlocks4k, Block4kCapacity - CountDelta);
      HeapFree(Heap, 0, Blocks4k);
      Blocks4k := NewBlocks4k;
      for I := CountDelta downto 1 do
        Blocks4k^[Block4kCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TBlock4k));
    end;
    with Blocks4k^[Block4kCount]^ do
    begin
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, $1000);
    Inc(Block4kCount);
  end;
end;

function G_GetMem8k: Pointer;
var
  I, Index: Integer;
  R: PRegion;
  NewBlocks8k: PBlock8kList;
begin
  Result := nil;
  for I := 0 to Block8kCount - 1 do
    with Blocks8k^[I]^ do
      if BitArray <> $FF then
      begin
        Index := G_FreeBitScanForward32(BitArray);
        BitArray := G_BitSet32(BitArray, Index);
        Result := Pointer(LongWord(Index shl 13) + BaseAddr);
        Inc(TotalMemAllocated, $2000);
        Exit;
      end;
  if MemFlags and mbfFreeBlock8kExists <> 0 then
    with Blocks8k^[Block8kCount]^ do
    begin
      BitArray := 1;
      Result := Pointer(BaseAddr);
      Inc(Block8kCount);
      MemFlags := MemFlags and not mbfFreeBlock8kExists;
      Inc(TotalMemAllocated, $2000);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitReset32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block8kCount = Block8kCapacity then
    begin
      Inc(Block8kCapacity, CountDelta);
      NewBlocks8k := HeapAlloc(Heap, 0, Block8kCapacity * SizeOf(Pointer));
      G_CopyLongs(Blocks8k, NewBlocks8k, Block8kCapacity - CountDelta);
      HeapFree(Heap, 0, Blocks8k);
      Blocks8k := NewBlocks8k;
      for I := CountDelta downto 1 do
        Blocks8k^[Block8kCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TBlock8k));
    end;
    with Blocks8k^[Block8kCount]^ do
    begin
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, $2000);
    Inc(Block8kCount);
  end;
end;

function G_GetMem16k: Pointer;
var
  I, Index: Integer;
  R: PRegion;
  NewBlocks16k: PBlock16kList;
begin
  Result := nil;
  for I := 0 to Block16kCount - 1 do
    with Blocks16k^[I]^ do
      if BitArray <> $F then
      begin
        Index := G_FreeBitScanForward32(BitArray);
        BitArray := G_BitSet32(BitArray, Index);
        Result := Pointer(LongWord(Index shl 14) + BaseAddr);
        Inc(TotalMemAllocated, $4000);
        Exit;
      end;
  if MemFlags and mbfFreeBlock16kExists <> 0 then
    with Blocks16k^[Block16kCount]^ do
    begin
      BitArray := 1;
      Result := Pointer(BaseAddr);
      Inc(Block16kCount);
      MemFlags := MemFlags and not mbfFreeBlock16kExists;
      Inc(TotalMemAllocated, $4000);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitReset32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block16kCount = Block16kCapacity then
    begin
      Inc(Block16kCapacity, CountDelta);
      NewBlocks16k := HeapAlloc(Heap, 0, Block16kCapacity * SizeOf(Pointer));
      G_CopyLongs(Blocks16k, NewBlocks16k, Block16kCapacity - CountDelta);
      HeapFree(Heap, 0, Blocks16k);
      Blocks16k := NewBlocks16k;
      for I := CountDelta downto 1 do
        Blocks16k^[Block16kCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TBlock16k));
    end;
    with Blocks16k^[Block16kCount]^ do
    begin
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, $4000);
    Inc(Block16kCount);
  end;
end;

function G_GetMem32k: Pointer;
var
  I: Integer;
  R: PRegion;
  NewBlocks32k: PBlock32kList;
begin
  Result := nil;
  for I := 0 to Block32kCount - 1 do
    with Blocks32k^[I]^ do
      if BitArray <> 3 then
      begin
        if BitArray and 1 = 0 then
        begin
          BitArray := BitArray or 1;
          Result := Pointer(BaseAddr);
        end else
        begin
          BitArray := BitArray or 2;
          Result := Pointer(BaseAddr + $8000);
        end;
        Inc(TotalMemAllocated, $8000);
        Exit;
      end;
  if MemFlags and mbfFreeBlock32kExists <> 0 then
    with Blocks32k^[Block32kCount]^ do
    begin
      BitArray := 1;
      Result := Pointer(BaseAddr);
      Inc(Block32kCount);
      MemFlags := MemFlags and not mbfFreeBlock32kExists;
      Inc(TotalMemAllocated, $8000);
      Exit;
    end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitReset32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block32kCount = Block32kCapacity then
    begin
      Inc(Block32kCapacity, CountDelta);
      NewBlocks32k := HeapAlloc(Heap, 0, Block32kCapacity * SizeOf(Pointer));
      G_CopyLongs(Blocks32k, NewBlocks32k, Block32kCapacity - CountDelta);
      HeapFree(Heap, 0, Blocks32k);
      Blocks32k := NewBlocks32k;
      for I := CountDelta downto 1 do
        Blocks32k^[Block32kCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TBlock32k));
    end;
    with Blocks32k^[Block32kCount]^ do
    begin
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      BitArray := 1;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, $8000);
    Inc(Block32kCount);
  end;
end;

function G_GetMem64k: Pointer;
var
  I: Integer;
  R: PRegion;
  NewBlocks64k: PBlock64kList;
begin
  Result := nil;
  if MemFlags and mbfFreeBlock64kExists <> 0 then
  begin
    Result := Pointer(Blocks64k^[Block64kCount].BaseAddr);
    Inc(Block64kCount);
    MemFlags := MemFlags and not mbfFreeBlock64kExists;
    Inc(TotalMemAllocated, $10000);
    Exit;
  end;
  R := G_GetRegion;
  if R <> nil then
  begin
    with R^ do
    begin
      ItemSize0 := G_BitSet32(ItemSize0, BlockIndex);
      ItemSize1 := G_BitSet32(ItemSize1, BlockIndex);
      ItemSize2 := G_BitSet32(ItemSize2, BlockIndex);
      ItemSize3 := G_BitSet32(ItemSize3, BlockIndex);
    end;
    if Block64kCount = Block64kCapacity then
    begin
      Inc(Block64kCapacity, CountDelta);
      NewBlocks64k := HeapAlloc(Heap, 0, Block64kCapacity * SizeOf(Pointer));
      G_CopyLongs(Blocks64k, NewBlocks64k, Block64kCapacity - CountDelta);
      HeapFree(Heap, 0, Blocks64k);
      Blocks64k := NewBlocks64k;
      for I := CountDelta downto 1 do
        Blocks64k^[Block64kCapacity - I] := HeapAlloc(Heap, 0, SizeOf(TBlock64k));
    end;
    with Blocks64k^[Block64kCount]^ do
    begin
      BaseAddr := LongWord(BlockIndex shl 16) + R.BaseAddr;
      Region := R;
      Result := Pointer(BaseAddr);
    end;
    Inc(TotalMemAllocated, $10000);
    Inc(Block64kCount);
  end;
end;

function G_GetBigMem(Size: Integer): Pointer;
var
  NewBigBlocks: PBigBlockList;
begin
  if BigBlockCount = BigBlockCapacity then
  begin
    Inc(BigBlockCapacity, CountDelta);
    NewBigBlocks := HeapAlloc(Heap, 0, BigBlockCapacity * SizeOf(Integer));
    G_CopyLongs(BigBlockAddrs, NewBigBlocks, BigBlockCapacity - CountDelta);
    HeapFree(Heap, 0, BigBlockAddrs);
    BigBlockAddrs := NewBigBlocks;
    NewBigBlocks := HeapAlloc(Heap, 0, BigBlockCapacity * SizeOf(Integer));
    G_CopyLongs(BigBlockSizes, NewBigBlocks, BigBlockCapacity - CountDelta);
    HeapFree(Heap, 0, BigBlockSizes);
    BigBlockSizes := NewBigBlocks;
  end;
  Result := VirtualAlloc(nil, Size, MEM_RESERVE or MEM_COMMIT, PAGE_READWRITE);
  if Result <> nil then
  begin
    BigBlockAddrs^[BigBlockCount] := Integer(Result);
    BigBlockSizes^[BigBlockCount] := Size;
    Inc(TotalMemAllocated, Size);
    Inc(BigBlockCount);
  end;
end;

function G_GetMem(Size: Integer): Pointer;
begin
  try
    if IsMultiThread then
      EnterCriticalSection(CriticalSection);
    if Size <= $400 then
    begin
      if Size <= 64 then
      begin
        if Size <= 16 then
        begin
          if Size <= 8 then
            Result := G_GetMem8
          else
            Result := G_GetMem16;
        end else
        begin
          if Size <= 32 then
            Result := G_GetMem32
          else
            Result := G_GetMem64;
        end;
      end else
      begin
        if Size <= 256 then
        begin
          if Size <= 128 then
            Result := G_GetMem128
          else
            Result := G_GetMem256;
        end else
        begin
          if Size <= 512 then
            Result := G_GetMem512
          else
            Result := G_GetMem1024;
        end;
      end;
    end else
    begin
      if Size <= $4000 then
      begin
        if Size <= $1000 then
        begin
          if Size <= $800 then
            Result := G_GetMem2k
          else
            Result := G_GetMem4k;
        end else
        begin
          if Size <= $2000 then
            Result := G_GetMem8k
          else
            Result := G_GetMem16k;
        end;
      end else
      begin
        if Size <= $10000 then
        begin
          if Size <= $8000 then
            Result := G_GetMem32k
          else
            Result := G_GetMem64k;
        end else
          Result := G_GetBigMem(((Size + $FFF) shr 12) shl 12);
      end;
    end;
  finally
    if IsMultiThread then
      LeaveCriticalSection(CriticalSection);
  end;
end;

procedure RaiseFreeUnusedBlockError;
begin
  raise Exception.Create(SErrFreeUnusedBlock);
end;

procedure G_FreeBlock(R: PRegion; BlockBaseAddr: LongWord);
var
  M: LongWord;
begin
  M := G_BitReset32(R.BlockMask, (BlockBaseAddr - R.BaseAddr) shr 16);
  R.BlockMask := M;
  if M = 0 then
  begin
    if MemFlags and mbfFreeRegionExists <> 0 then
      VirtualFree(Pointer(Regions^[RegionCount].BaseAddr), 0, MEM_RELEASE)
    else
      MemFlags := MemFlags or mbfFreeRegionExists;
    Dec(RegionCount);
    if Regions^[RegionCount] <> R then
    begin
      Regions^[G_Scan_Pointer(R, Regions, RegionCount)] := Regions^[RegionCount];
      Regions^[RegionCount] := R;
    end;
  end;
end;

procedure G_FreeMem8(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 8);
  for I := Block8Count - 1 downto 0 do
    with Blocks8^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 8192 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock8Exists <> 0 then
    with Blocks8^[Block8Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock8Exists;
  Dec(Block8Count);
  if J < Block8Count then
  begin
    I := Integer(Blocks8^[J]);
    Blocks8^[J] := Blocks8^[Block8Count];
    Blocks8^[Block8Count] := Pointer(I);
  end;
end;

procedure G_FreeMem16(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 16);
  for I := Block16Count - 1 downto 0 do
    with Blocks16^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 4096 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock16Exists <> 0 then
    with Blocks16^[Block16Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock16Exists;
  Dec(Block16Count);
  if J < Block16Count then
  begin
    I := Integer(Blocks16^[J]);
    Blocks16^[J] := Blocks16^[Block16Count];
    Blocks16^[Block16Count] := Pointer(I);
  end;
end;

procedure G_FreeMem32(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 32);
  for I := Block32Count - 1 downto 0 do
    with Blocks32^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 2048 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock32Exists <> 0 then
    with Blocks32^[Block32Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock32Exists;
  Dec(Block32Count);
  if J < Block32Count then
  begin
    I := Integer(Blocks32^[J]);
    Blocks32^[J] := Blocks32^[Block32Count];
    Blocks32^[Block32Count] := Pointer(I);
  end;
end;

procedure G_FreeMem64(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 64);
  for I := Block64Count - 1 downto 0 do
    with Blocks64^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 1024 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock64Exists <> 0 then
    with Blocks64^[Block64Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock64Exists;
  Dec(Block64Count);
  if J < Block64Count then
  begin
    I := Integer(Blocks64^[J]);
    Blocks64^[J] := Blocks64^[Block64Count];
    Blocks64^[Block64Count] := Pointer(I);
  end;
end;

procedure G_FreeMem128(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 128);
  for I := Block128Count - 1 downto 0 do
    with Blocks128^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 512 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock128Exists <> 0 then
    with Blocks128^[Block128Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock128Exists;
  Dec(Block128Count);
  if J < Block128Count then
  begin
    I := Integer(Blocks128^[J]);
    Blocks128^[J] := Blocks128^[Block128Count];
    Blocks128^[Block128Count] := Pointer(I);
  end;
end;

procedure G_FreeMem256(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 256);
  for I := Block256Count - 1 downto 0 do
    with Blocks256^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 256 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock256Exists <> 0 then
    with Blocks256^[Block256Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock256Exists;
  Dec(Block256Count);
  if J < Block256Count then
  begin
    I := Integer(Blocks256^[J]);
    Blocks256^[J] := Blocks256^[Block256Count];
    Blocks256^[Block256Count] := Pointer(I);
  end;
end;

procedure G_FreeMem512(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 512);
  for I := Block512Count - 1 downto 0 do
    with Blocks512^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 128 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock512Exists <> 0 then
    with Blocks512^[Block512Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock512Exists;
  Dec(Block512Count);
  if J < Block512Count then
  begin
    I := Integer(Blocks512^[J]);
    Blocks512^[J] := Blocks512^[Block512Count];
    Blocks512^[Block512Count] := Pointer(I);
  end;
end;

procedure G_FreeMem1024(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, 1024);
  for I := Block1024Count - 1 downto 0 do
    with Blocks1024^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Index shr 5;
        if LongWord(J) < LongWord(FirstFree) then
          FirstFree := J;
        Inc(FreeCount);
        if not G_BitReset(@BitArray, Index) then
          RaiseFreeUnusedBlockError;
        if FreeCount < 64 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock1024Exists <> 0 then
    with Blocks1024^[Block1024Count]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock1024Exists;
  Dec(Block1024Count);
  if J < Block1024Count then
  begin
    I := Integer(Blocks1024^[J]);
    Blocks1024^[J] := Blocks1024^[Block1024Count];
    Blocks1024^[Block1024Count] := Pointer(I);
  end;
end;

procedure G_FreeMem2k(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, $800);
  for I := Block2kCount - 1 downto 0 do
    with Blocks2k^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Integer(BitArray);
        BitArray := G_BitReset32(BitArray, Index);
        if J = Integer(BitArray) then
          RaiseFreeUnusedBlockError;
        if BitArray <> 0 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock2kExists <> 0 then
    with Blocks2k^[Block2kCount]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock2kExists;
  Dec(Block2kCount);
  if J < Block2kCount then
  begin
    I := Integer(Blocks2k^[J]);
    Blocks2k^[J] := Blocks2k^[Block2kCount];
    Blocks2k^[Block2kCount] := Pointer(I);
  end;
end;

procedure G_FreeMem4k(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, $1000);
  for I := Block4kCount - 1 downto 0 do
    with Blocks4k^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Integer(BitArray);
        BitArray := G_BitReset32(BitArray, Index);
        if J = Integer(BitArray) then
          RaiseFreeUnusedBlockError;
        if BitArray <> 0 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock4kExists <> 0 then
    with Blocks4k^[Block4kCount]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock4kExists;
  Dec(Block4kCount);
  if J < Block4kCount then
  begin
    I := Integer(Blocks4k^[J]);
    Blocks4k^[J] := Blocks4k^[Block4kCount];
    Blocks4k^[Block4kCount] := Pointer(I);
  end;
end;

procedure G_FreeMem8k(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, $2000);
  for I := Block8kCount - 1 downto 0 do
    with Blocks8k^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Integer(BitArray);
        BitArray := G_BitReset32(BitArray, Index);
        if J = Integer(BitArray) then
          RaiseFreeUnusedBlockError;
        if BitArray <> 0 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock8kExists <> 0 then
    with Blocks8k^[Block8kCount]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock8kExists;
  Dec(Block8kCount);
  if J < Block8kCount then
  begin
    I := Integer(Blocks8k^[J]);
    Blocks8k^[J] := Blocks8k^[Block8kCount];
    Blocks8k^[Block8kCount] := Pointer(I);
  end;
end;

procedure G_FreeMem16k(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, $4000);
  for I := Block16kCount - 1 downto 0 do
    with Blocks16k^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Integer(BitArray);
        BitArray := G_BitReset32(BitArray, Index);
        if J = Integer(BitArray) then
          RaiseFreeUnusedBlockError;
        if BitArray <> 0 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock16kExists <> 0 then
    with Blocks16k^[Block16kCount]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock16kExists;
  Dec(Block16kCount);
  if J < Block16kCount then
  begin
    I := Integer(Blocks16k^[J]);
    Blocks16k^[J] := Blocks16k^[Block16kCount];
    Blocks16k^[Block16kCount] := Pointer(I);
  end;
end;

procedure G_FreeMem32k(BlockBaseAddr: LongWord; Index: Integer);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, $8000);
  for I := Block32kCount - 1 downto 0 do
    with Blocks32k^[I]^ do
      if BaseAddr = BlockBaseAddr then
      begin
        J := Integer(BitArray);
        BitArray := G_BitReset32(BitArray, Index);
        if J = Integer(BitArray) then
          RaiseFreeUnusedBlockError;
        if BitArray <> 0 then
          Exit;
        J := I;
        Break;
      end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock32kExists <> 0 then
    with Blocks32k^[Block32kCount]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock32kExists;
  Dec(Block32kCount);
  if J < Block32kCount then
  begin
    I := Integer(Blocks32k^[J]);
    Blocks32k^[J] := Blocks32k^[Block32kCount];
    Blocks32k^[Block32kCount] := Pointer(I);
  end;
end;

procedure G_FreeMem64k(BlockBaseAddr: LongWord);
var
  I, J: Integer;
begin
  J := -1;
  Dec(TotalMemAllocated, $10000);
  for I := Block64kCount - 1 downto 0 do
    if Blocks64k^[I].BaseAddr = BlockBaseAddr then
    begin
      J := I;
      Break;
    end;
  if J < 0 then
    RaiseFreeUnusedBlockError;
  if MemFlags and mbfFreeBlock64kExists <> 0 then
    with Blocks64k^[Block64kCount]^ do
      G_FreeBlock(Region, BaseAddr)
  else
    MemFlags := MemFlags or mbfFreeBlock64kExists;
  Dec(Block64kCount);
  if J < Block64kCount then
  begin
    I := Integer(Blocks64k^[J]);
    Blocks64k^[J] := Blocks64k^[Block64kCount];
    Blocks64k^[Block64kCount] := Pointer(I);
  end;
end;

procedure G_FreeBigMem(BigBlockIndex: Integer);
begin
  Dec(TotalMemAllocated, BigBlockSizes^[BigBlockIndex]);
  VirtualFree(Pointer(BigBlockAddrs^[BigBlockIndex]), 0, MEM_RELEASE);
  Dec(BigBlockCount);
  if BigBlockIndex < BigBlockCount then
  begin
    BigBlockAddrs^[BigBlockIndex] := BigBlockAddrs^[BigBlockCount];
    BigBlockSizes^[BigBlockIndex] := BigBlockSizes^[BigBlockCount];
  end;
end;

function G_FreeMem(P: Pointer): Integer;
var
  I: Integer;
  L, N, K: LongWord;
  R: PRegion;
  Found: Boolean;
begin
  try
    if IsMultiThread then
      EnterCriticalSection(CriticalSection);
    Result := 0;
    R := nil;
    L := LongWord(P);
    N := 0;
    K := 0;
    Found := False;
    for I := RegionCount - 1 downto 0 do
    begin
      R := Regions^[I];
      K := R.BaseAddr;
      N := L - K;
      if N < $200000 then
      begin
        Found := True;
        Break;
      end;
    end;
    if not Found then
    begin
      I := G_ScanBackward_LongWord(L, BigBlockAddrs, BigBlockCount);
      if I >= 0 then
        G_FreeBigMem(I)
      else
        Result := OldMemoryManager.FreeMem(Pointer(L));
    end else
    begin
      I := Integer(N and $FFFF);
      Inc(K, N and $001F0000);
      N := BitMasks32[N shr 16];
      if R.ItemSize0 and N = 0 then
      begin
        if R.ItemSize1 and N = 0 then
        begin
          if R.ItemSize2 and N = 0 then
            G_FreeMem512(K, I shr 9)
          else if R.ItemSize3 and N = 0 then
            G_FreeMem32(K, I shr 5)
          else
            G_FreeMem8k(K, I shr 13);
        end else
        begin
          if R.ItemSize2 and N = 0 then
          begin
            if R.ItemSize3 and N = 0 then
              G_FreeMem8(K, I shr 3)
            else
              G_FreeMem2k(K, I shr 11);
          end else
          begin
            if R.ItemSize3 and N = 0 then
              G_FreeMem128(K, I shr 7)
            else
              G_FreeMem32k(K, I shr 15);
          end;
        end;
      end else
      begin
        if R.ItemSize1 and N = 0 then
        begin
          if R.ItemSize2 and N = 0 then
            G_FreeMem1024(K, I shr 10)
          else if R.ItemSize3 and N = 0 then
            G_FreeMem64(K, I shr 6)
          else
            G_FreeMem16k(K, I shr 14);
        end else
        begin
          if R.ItemSize2 and N = 0 then
          begin
            if R.ItemSize3 and N = 0 then
              G_FreeMem16(K, I shr 4)
            else
              G_FreeMem4k(K, I shr 12);
          end else
          begin
            if R.ItemSize3 and N = 0 then
              G_FreeMem256(K, I shr 8)
            else
              G_FreeMem64k(K);
          end;
        end;
      end;
    end;
  finally
    if IsMultiThread then
      LeaveCriticalSection(CriticalSection);
  end;
end;

function G_ReallocMem(P: Pointer; Size: Integer): Pointer;
var
  L: Integer;
  Mem: Pointer;
begin
  L := G_GetMemSize(P);
  if L = 0 then
  begin
    Result := OldMemoryManager.ReallocMem(P, Size);
    Exit;
  end;
  if Size <= $400 then
  begin
    if Size <= 64 then
    begin
      if Size <= 16 then
      begin
        if Size <= 8 then
          Size := 8
        else
          Size := 16;
      end else
      begin
        if Size <= 32 then
          Size := 32
        else
          Size := 64;
      end;
    end else
    begin
      if Size <= 256 then
      begin
        if Size <= 128 then
          Size := 128
        else
          Size := 256;
      end else
      begin
        if Size <= 512 then
          Size := 512
        else
          Size := 1024;
      end;
    end;
  end else
  begin
    if Size <= $4000 then
    begin
      if Size <= $1000 then
      begin
        if Size <= $800 then
          Size := $800
        else
          Size := $1000;
      end else
      begin
        if Size <= $2000 then
          Size := $2000
        else
          Size := $4000;
      end;
    end else
    begin
      if Size <= $10000 then
      begin
        if Size <= $8000 then
          Size := $8000
        else
          Size := $10000;
      end else
        Size := ((Size + $FFF) shr 12) shl 12;
    end;
  end;
  if Size = L then
    Result := P
  else
  begin
    Mem := G_GetMem(Size);
    if L < Size then
      Size := L;
    G_CopyLongs(P, Mem, Size shr 2);
    G_FreeMem(P);
    Result := Mem;
  end;
end;

function G_GetMemSize(P: Pointer): Integer;
var
  I: Integer;
  L, N: LongWord;
  R: PRegion;
  Found: Boolean;
begin
  if IsMultiThread then
    EnterCriticalSection(CriticalSection);
  Result := 0;
  L := LongWord(P);
  if L <> 0 then
  begin
    R := nil;
    N := 0;
    Found := False;
    for I := RegionCount - 1 downto 0 do
    begin
      R := Regions^[I];
      N := L - R.BaseAddr;
      if N < $200000 then
      begin
        Found := True;
        Break;
      end;
    end;
    if not Found then
    begin
      I := G_ScanBackward_LongWord(L, BigBlockAddrs, BigBlockCount);
      if I >= 0 then
        Result := BigBlockSizes^[I];
    end else
    begin
      N := BitMasks32[N shr 16];
      if R.ItemSize0 and N = 0 then
      begin
        if R.ItemSize1 and N = 0 then
        begin
          if R.ItemSize2 and N = 0 then
            Result := 512
          else if R.ItemSize3 and N = 0 then
            Result := 32
          else
            Result := 8192;
        end else
        begin
          if R.ItemSize2 and N = 0 then
          begin
            if R.ItemSize3 and N = 0 then
              Result := 8
            else
              Result := 2048;
          end else
          begin
            if R.ItemSize3 and N = 0 then
              Result := 128
            else
              Result := 32768;
          end;
        end;
      end else
      begin
        if R.ItemSize1 and N = 0 then
        begin
          if R.ItemSize2 and N = 0 then
            Result := 1024
          else if R.ItemSize3 and N = 0 then
            Result := 64
          else
            Result := 16384;
        end else
        begin
          if R.ItemSize2 and N = 0 then
          begin
            if R.ItemSize3 and N = 0 then
              Result := 16
            else
              Result := 4096;
          end else
          begin
            if R.ItemSize3 and N = 0 then
              Result := 256
            else
              Result := 65536;
          end;
        end;
      end;
    end;
  end;
  if IsMultiThread then
    LeaveCriticalSection(CriticalSection);
end;

function G_GetTotalMemAllocated: LongWord;
begin
  if IsMultiThread then
    EnterCriticalSection(CriticalSection);
  Result := TotalMemAllocated;
  if IsMultiThread then
    LeaveCriticalSection(CriticalSection);
end;

function G_GetTotalMemCommitted: LongWord;
var
  I: Integer;
begin
  if IsMultiThread then
    EnterCriticalSection(CriticalSection);
  Result := RegionCount;
  if MemFlags and mbfFreeRegionExists <> 0 then
    Inc(Result);
  Result := Result shl 21;
  for I := BigBlockCount - 1 downto 0 do
    Inc(Result, BigBlockSizes^[I]);
  if IsMultiThread then
    LeaveCriticalSection(CriticalSection);
end;

const
  G_Manager: TMemoryManager = (GetMem: G_GetMem; FreeMem: G_FreeMem; ReallocMem: G_ReallocMem);

procedure InitHeap;
var
  I: Integer;
begin
  if MemFlags and mbfManagerInstalled <> 0 then
    Exit;
  Heap := HeapCreate(winHEAP_NO_SERIALIZE or winHEAP_GENERATE_EXCEPTIONS, $20000, $800000);
  RegionCapacity := CountDelta;
  Regions := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Regions^[I] := HeapAlloc(Heap, 0, SizeOf(TRegion));
  Blocks8 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks8^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock8));
  Block8Capacity := CountDelta;
  Blocks16 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks16^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock16));
  Block16Capacity := CountDelta;
  Blocks32 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks32^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock32));
  Block32Capacity := CountDelta;
  Blocks64 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks64^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock64));
  Block64Capacity := CountDelta;
  Blocks128 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks128^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock128));
  Block128Capacity := CountDelta;
  Blocks256 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks256^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock256));
  Block256Capacity := CountDelta;
  Blocks512 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks512^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock512));
  Block512Capacity := CountDelta;
  Blocks1024 := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks1024^[I] := HeapAlloc(Heap, winHEAP_ZERO_MEMORY, SizeOf(TBlock1024));
  Block1024Capacity := CountDelta;
  Blocks2k := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks2k^[I] := HeapAlloc(Heap, 0, SizeOf(TBlock2k));
  Block2kCapacity := CountDelta;
  Blocks4k := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks4k^[I] := HeapAlloc(Heap, 0, SizeOf(TBlock4k));
  Block4kCapacity := CountDelta;
  Blocks8k := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks8k^[I] := HeapAlloc(Heap, 0, SizeOf(TBlock8k));
  Block8kCapacity := CountDelta;
  Blocks16k := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks16k^[I] := HeapAlloc(Heap, 0, SizeOf(TBlock16k));
  Block16kCapacity := CountDelta;
  Blocks32k := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks32k^[I] := HeapAlloc(Heap, 0, SizeOf(TBlock32k));
  Block32kCapacity := CountDelta;
  Blocks64k := HeapAlloc(Heap, 0, CountDelta * SizeOf(Pointer));
  for I := 0 to CountDelta - 1 do
    Blocks64k^[I] := HeapAlloc(Heap, 0, SizeOf(TBlock64k));
  Block64kCapacity := CountDelta;
  BigBlockAddrs := HeapAlloc(Heap, 0, CountDelta * SizeOf(Integer));
  BigBlockSizes := HeapAlloc(Heap, 0, CountDelta * SizeOf(Integer));
  BigBlockCapacity := CountDelta;
  with Regions^[0]^ do
  begin
    BaseAddr := LongWord(VirtualAlloc(nil, $200000,
      MEM_RESERVE or MEM_COMMIT or MEM_TOP_DOWN, PAGE_READWRITE));
    BlockMask := 0;
  end;
  RegionCount := 1;
  InitializeCriticalSection(CriticalSection);
  GetMemoryManager(OldMemoryManager);
  MemFlags := MemFlags or mbfManagerInstalled;
  TotalMemAllocated := 0;
  SetMemoryManager(G_Manager);
end;

procedure DoneHeap;
var
  I: Integer;
begin
  if MemFlags and mbfManagerInstalled = 0 then
    Exit;
  SetMemoryManager(OldMemoryManager);
  MemFlags := MemFlags and not mbfManagerInstalled;
  DeleteCriticalSection(CriticalSection);
  for I := BigBlockCount - 1 downto 0 do
    VirtualFree(Pointer(BigBlockAddrs^[I]), 0, MEM_RELEASE);
  if MemFlags and mbfFreeRegionExists <> 0 then
    VirtualFree(Pointer(Regions^[RegionCount].BaseAddr), 0, MEM_RELEASE);
  for I := RegionCount - 1 downto 0 do
    VirtualFree(Pointer(Regions^[I].BaseAddr), 0, MEM_RELEASE);
  HeapDestroy(Heap);
end;

initialization
  InitHeap;

finalization
  if not IsMultiThread then
    DoneHeap;

end.

